package com.trafficparrot.sdk.file;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.google.common.base.Function;
import com.google.common.base.Predicate;
import com.google.common.collect.ComparisonChain;
import com.trafficparrot.sdk.jms.Destination;
import com.trafficparrot.sdk.jms.DestinationType;
import org.apache.commons.lang3.StringUtils;

import static com.trafficparrot.sdk.jms.DestinationType.QUEUE;
import static com.trafficparrot.sdk.jms.DestinationType.TOPIC;
import static java.lang.String.format;
import static org.apache.commons.lang3.builder.ToStringBuilder.reflectionToString;
import static org.apache.commons.lang3.builder.ToStringStyle.SHORT_PREFIX_STYLE;

public class FileMessageMapping implements Comparable<FileMessageMapping> {

    public final String mappingId;
    public final FileMessageRequestMatcher request;
    public final FileMessageResponse response;

    public FileMessageMapping(@JsonProperty("mappingId") String mappingId,
                              @JsonProperty("request") FileMessageRequestMatcher request,
                              @JsonProperty("response") FileMessageResponse response) {
        this.mappingId = mappingId;
        this.request = request;
        this.response = response;
    }
    
    @JsonIgnore
    public FileMessageResponse transformedResponse(FileMessage requestMessage, FileMessageHandlebarsResponseTransformer handlebars) {
        FileMessageResponse afterResponseTransformer = responseTransformerTransform(requestMessage);
        FileMessageResponse afterHandlebars = handlebars.transform(requestMessage, afterResponseTransformer);
        return afterHandlebars;
    }

    private FileMessageResponse responseTransformerTransform(FileMessage requestMessage) {
        if (StringUtils.isNotBlank(response.fileMessageResponseTransformerClassName) && !FileMessageResponse.NO_TRANSFORMER.equals(response.fileMessageResponseTransformerClassName)) {
            FileMessageResponseTransformer jmsResponseTransformer = parseExtension(response.fileMessageResponseTransformerClassName);
            return jmsResponseTransformer.transform(request.destination, requestMessage, response);
        } else {
            return response;
        }
    }

    @JsonIgnore
    private static FileMessageResponseTransformer parseExtension(String extensionsPropertyValue) {
        try {
            return Class.forName(extensionsPropertyValue).asSubclass(FileMessageResponseTransformer.class).newInstance();
        } catch (ClassNotFoundException e) {
            throw new IllegalArgumentException(format("Extension '%s' was not found on the classpath", extensionsPropertyValue), e);
        } catch (IllegalAccessException | InstantiationException e) {
            throw new IllegalArgumentException(format("Extension '%s' must have a public zero argument constructor", extensionsPropertyValue), e);
        }
    }

    @Override
    @JsonIgnore
    public String toString() {
        return reflectionToString(this, SHORT_PREFIX_STYLE);
    }

    @Override
    public int compareTo(FileMessageMapping other) {
        return ComparisonChain.start()
                .compare(mappingId, other.mappingId)
                .compare(request, other.request)
                .compare(response, other.response)
                .result();
    }

    public static Predicate<FileMessageMapping> isQueue() {
        return new Predicate<FileMessageMapping>() {
            @Override
            public boolean apply(FileMessageMapping input) {
                return QUEUE.equals(input.request.destination.type);
            }
        };
    }

    public static Predicate<FileMessageMapping> isTopic() {
        return new Predicate<FileMessageMapping>() {
            @Override
            public boolean apply(FileMessageMapping input) {
                return TOPIC.equals(input.request.destination.type);
            }
        };
    }

    public static Function<FileMessageMapping, String> destinationName() {
        return new Function<FileMessageMapping, String>() {
            @Override
            public String apply(FileMessageMapping input) {
                return input.request.destination.name;
            }
        };
    }

    public static Function<FileMessageMapping, Destination> destination() {
        return new Function<FileMessageMapping, Destination>() {
            @Override
            public Destination apply(FileMessageMapping input) {
                return input.request.destination;
            }
        };
    }

    public static Predicate<? super FileMessageMapping> withDestination(final Destination destination) {
        return new Predicate<FileMessageMapping>() {
            @Override
            public boolean apply(FileMessageMapping input) {
                return input.request.destination.equals(destination);
            }
        };
    }

    public static Predicate<? super FileMessageMapping> withDestinationType(final DestinationType type) {
        return new Predicate<FileMessageMapping>() {
            @Override
            public boolean apply(FileMessageMapping input) {
                return input.request.destination.type == type;
            }
        };
    }
}
