package com.trafficparrot.sdk.ibmmq;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.ComparisonChain;
import com.ibm.mq.MQMessage;
import com.trafficparrot.sdk.jms.Destination;

import java.io.IOException;

import static com.trafficparrot.sdk.ibmmq.IbmMqMessageFormat.MQFMT_STRING;
import static com.trafficparrot.sdk.messaging.PrintableCharactersRepresentation.decodePrintableCharacters;
import static org.apache.commons.lang3.ObjectUtils.firstNonNull;
import static org.apache.commons.lang3.builder.ToStringBuilder.reflectionToString;
import static org.apache.commons.lang3.builder.ToStringStyle.SHORT_PREFIX_STYLE;

public class IbmMqResponse implements Comparable<IbmMqResponse> {
    public static final String NO_TRANSFORMER = "NO_TRANSFORMER";

    public final Destination destination;
    public final String ibmMqResponseTransformerClassName;
    public final IbmMqMessageFormat format;
    public final String text;

    public IbmMqResponse(@JsonProperty("destination") Destination destination,
                         @JsonProperty("ibmMqResponseTransformerClassName") String ibmMqResponseTransformerClassName,
                         @JsonProperty("format") IbmMqMessageFormat format,
                         @JsonProperty("text") String text) {
        this.destination = destination;
        this.ibmMqResponseTransformerClassName = ibmMqResponseTransformerClassName;
        this.format = firstNonNull(format, MQFMT_STRING);
        this.text = text;
    }

    public IbmMqResponse(Destination destination, String text) {
        this(destination, NO_TRANSFORMER, MQFMT_STRING, text);
    }

    public MQMessage createMessage() throws IOException {
        MQMessage mqMessage = new MQMessage();
        writeMessage(mqMessage);
        return mqMessage;
    }

    @VisibleForTesting
    void writeMessage(MQMessage mqMessage) throws IOException {
        mqMessage.format = format.value;
        switch (format) {
            case MQFMT_RF_HEADER_1: // header is binary
            case MQFMT_RF_HEADER_2: // header is binary
            case MQFMT_NONE: // might include binary data
                mqMessage.write(decodePrintableCharacters(text));
                break;
            case MQFMT_STRING:
                mqMessage.writeString(text);
                break;
        }
    }

    @Override
    @JsonIgnore
    public String toString() {
        return reflectionToString(this, SHORT_PREFIX_STYLE);
    }

    @Override
    public int compareTo(IbmMqResponse other) {
        return ComparisonChain.start()
                .compare(destination, other.destination)
                .compare(text, other.text)
                .result();
    }
}
