package com.trafficparrot.sdk.messaging;

import com.google.common.annotations.VisibleForTesting;

import java.util.List;

import static com.google.common.primitives.Bytes.concat;
import static com.trafficparrot.sdk.messaging.FindAndReplaceBytes.findOccurrences;
import static java.lang.Character.isLetterOrDigit;
import static java.lang.Character.isWhitespace;
import static java.nio.charset.StandardCharsets.UTF_8;
import static java.util.Arrays.copyOfRange;
import static org.apache.commons.lang3.CharUtils.isAsciiPrintable;

// TODO(3/5) if the encoding approach here gives us problems, try using ISO_8859_1 instead and get rid of the custom escaping
public class PrintableCharactersRepresentation {

    private static final String NEGATIVE_BYTE_ESCAPE = "\b-";
    private static final byte[] NEGATIVE_BYTE_ESCAPE_BYTES = NEGATIVE_BYTE_ESCAPE.getBytes(UTF_8);
    private static final int MAKE_POSITIVE_OFFSET = 128;

    public static String encodePrintableCharacters(byte[] bytesToEncode) {
        byte[] encoded = new byte[bytesToEncode.length + numberOfNegativeBytes(bytesToEncode) * NEGATIVE_BYTE_ESCAPE_BYTES.length];
        int position = 0;
        for (byte toEncode : bytesToEncode) {
            if (toEncode >= 0) {
                encoded[position++] = toEncode;
            } else {
                for (byte escape : NEGATIVE_BYTE_ESCAPE_BYTES) {
                    encoded[position++] = escape;
                }
                encoded[position++] = (byte) (toEncode + MAKE_POSITIVE_OFFSET);
            }
        }
        return new String(encoded, UTF_8);
    }

    public static byte[] decodePrintableCharacters(String text) {
        byte[] toDecode = text.getBytes(UTF_8);
        List<Integer> occurrences = findOccurrences(toDecode, NEGATIVE_BYTE_ESCAPE_BYTES);
        byte[][] pieces = new byte[occurrences.size() * 2 + 1][];
        int position = 0;
        for (int i = 0; i < occurrences.size(); i++) {
            int occurrence = occurrences.get(i);
            byte[] before = copyOfRange(toDecode, position, occurrence);
            pieces[2 * i] = before;
            byte b = toDecode[occurrence + NEGATIVE_BYTE_ESCAPE_BYTES.length];
            pieces[2 * i + 1] = new byte[]{(byte) (b - MAKE_POSITIVE_OFFSET)};
            position = occurrence + NEGATIVE_BYTE_ESCAPE_BYTES.length + 1;
        }
        byte[] after = copyOfRange(toDecode, position, toDecode.length);
        pieces[occurrences.size() * 2] = after;
        return concat(pieces);
    }

    public static boolean isMostlyPrintableCharacters(byte[] bytes, double mostlyPrintableThreshold) {
        return (double) numberOfPrintableCharacters(bytes) / bytes.length > mostlyPrintableThreshold;
    }

    @VisibleForTesting
    static int numberOfPrintableCharacters(byte[] bytes) {
        String string = new String(bytes, UTF_8);
        int count = 0;
        for (int i = string.length() - 1; i >= 0; i--) {
            char character = string.charAt(i);
            if (isAsciiPrintable(character) || isLetterOrDigit(character) || isWhitespace(character)) {
                count++;
            }
        }
        return count;
    }

    private static int numberOfNegativeBytes(byte[] bytes) {
        int count = 0;
        for (byte aByte : bytes) {
            if (aByte < 0) {
                count++;
            }
        }
        return count;
    }
}
